@@ -1,28 +1,33 @@
 """Kazoo Zookeeper Client"""
 import inspect
 import logging
+import os
 import re
-from collections import defaultdict
+import warnings
+from collections import defaultdict, deque
 from functools import partial
 from os.path import split
 
 from kazoo.exceptions import (
     AuthFailedError,
-    ConnectionLoss,
+    ConfigurationError,
     ConnectionClosedError,
+    ConnectionLoss,
     NoNodeError,
     NodeExistsError,
-    ConfigurationError,
-    SessionExpiredError
+    SessionExpiredError,
+    WriterNotClosedException,
 )
 from kazoo.handlers.threading import SequentialThreadingHandler
+from kazoo.handlers.utils import capture_exceptions, wrap
 from kazoo.hosts import collect_hosts
 from kazoo.protocol.connection import ConnectionHandler
 from kazoo.protocol.paths import normpath
 from kazoo.protocol.paths import _prefix_root
 from kazoo.protocol.serialization import (
     Auth,
-    Close,
+    CheckVersion,
+    CloseInstance,
     Create,
     Delete,
     Exists,
@@ -32,7 +37,8 @@ from kazoo.protocol.serialization import (
     SetACL,
     GetData,
     SetData,
-    Sync
+    Sync,
+    Transaction
 )
 from kazoo.protocol.states import KazooState
 from kazoo.protocol.states import KeeperState
@@ -40,12 +46,49 @@ from kazoo.retry import KazooRetry
 from kazoo.security import ACL
 from kazoo.security import OPEN_ACL_UNSAFE
 
+# convenience API
+from kazoo.recipe.barrier import Barrier
+from kazoo.recipe.barrier import DoubleBarrier
+from kazoo.recipe.counter import Counter
+from kazoo.recipe.election import Election
+from kazoo.recipe.lock import Lock
+from kazoo.recipe.lock import Semaphore
+from kazoo.recipe.partitioner import SetPartitioner
+from kazoo.recipe.party import Party
+from kazoo.recipe.party import ShallowParty
+from kazoo.recipe.queue import Queue
+from kazoo.recipe.queue import LockingQueue
+from kazoo.recipe.watchers import ChildrenWatch
+from kazoo.recipe.watchers import DataWatch
+
+try:  # pragma: nocover
+    basestring
+except NameError:  # pragma: nocover
+    basestring = str
+
 LOST_STATES = (KeeperState.EXPIRED_SESSION, KeeperState.AUTH_FAILED,
                KeeperState.CLOSED)
 ENVI_VERSION = re.compile('[\w\s:.]*=([\d\.]*).*', re.DOTALL)
 log = logging.getLogger(__name__)
 
 
+_RETRY_COMPAT_DEFAULTS = dict(
+    max_retries=None,
+    retry_delay=0.1,
+    retry_backoff=2,
+    retry_jitter=0.8,
+    retry_max_delay=3600,
+)
+
+_RETRY_COMPAT_MAPPING = dict(
+    max_retries='max_tries',
+    retry_delay='delay',
+    retry_backoff='backoff',
+    retry_jitter='max_jitter',
+    retry_max_delay='max_delay',
+)
+
+
 class KazooClient(object):
     """An Apache Zookeeper Python client supporting alternate callback
     handlers and high-level functionality.
@@ -56,31 +99,19 @@ class KazooClient(object):
     :class:`~kazoo.protocol.states.WatchedEvent` instance.
 
     """
-    def __init__(self, hosts='127.0.0.1:2181', watcher=None,
-                 timeout=10.0, client_id=None, max_retries=None, retry_delay=0.1,
-                 retry_backoff=2, retry_jitter=0.8, handler=None,
-                 default_acl=None, auth_data=None, read_only=None):
+    def __init__(self, hosts='127.0.0.1:2181',
+                 timeout=10.0, client_id=None, handler=None,
+                 default_acl=None, auth_data=None, read_only=None,
+                 randomize_hosts=True, connection_retry=None,
+                 command_retry=None, logger=None, **kwargs):
         """Create a :class:`KazooClient` instance. All time arguments
         are in seconds.
 
         :param hosts: Comma-separated list of hosts to connect to
                       (e.g. 127.0.0.1:2181,127.0.0.1:2182).
-        :param watcher:
-            Set a default watcher. This will be called by the actual
-            default watcher that :class:`KazooClient` establishes.
         :param timeout: The longest to wait for a Zookeeper connection.
         :param client_id: A Zookeeper client id, used when
                           re-establishing a prior session connection.
-        :param max_retries: Maximum retries when using the
-                            :meth:`KazooClient.retry` method.
-        :param retry_delay: Initial delay when retrying a call.
-        :param retry_backoff:
-            Backoff multiplier between retry attempts. Defaults to 2
-            for exponential back-off.
-        :param retry_jitter:
-            How much jitter delay to introduce per call. An amount of
-            time up to this will be added per retry call to avoid
-            hammering the server.
         :param handler: An instance of a class implementing the
                         :class:`~kazoo.interfaces.IHandler` interface
                         for callback handling.
@@ -89,10 +120,18 @@ class KazooClient(object):
             A list of authentication credentials to use for the
             connection. Should be a list of (scheme, credential)
             tuples as :meth:`add_auth` takes.
-
-        Retry parameters will be used for connection establishment
-        attempts and reconnects.
-
+        :param read_only: Allow connections to read only servers.
+        :param randomize_hosts: By default randomize host selection.
+        :param connection_retry:
+            A :class:`kazoo.retry.KazooRetry` object to use for
+            retrying the connection to Zookeeper. Also can be a dict of
+            options which will be used for creating one.
+        :param command_retry:
+            A :class:`kazoo.retry.KazooRetry` object to use for
+            the :meth:`KazooClient.retry` method. Also can be a dict of
+            options which will be used for creating one.
+        :param logger: A custom logger to use instead of the module
+            global `log` instance.
 
         Basic Example:
 
@@ -110,8 +149,23 @@ class KazooClient(object):
             zk.start()
             lock = zk.Lock('/lock_path')
 
+        .. versionadded:: 0.6
+            The read_only option. Requires Zookeeper 3.4+
+
+        .. versionadded:: 0.6
+            The retry_max_delay option.
+
+        .. versionadded:: 0.6
+            The randomize_hosts option.
+
+        .. versionchanged:: 0.8
+            Removed the unused watcher argument (was second argument).
+
+        .. versionadded:: 1.2
+            The connection_retry, command_retry and logger options.
+
         """
-        self.log_debug = logging.DEBUG >= log.getEffectiveLevel()
+        self.logger = logger or log
 
         # Record the handler strategy used
         self.handler = handler if handler else SequentialThreadingHandler()
@@ -121,7 +175,8 @@ class KazooClient(object):
 
         self.auth_data = auth_data if auth_data else set([])
         self.default_acl = default_acl
-        self.hosts, chroot = collect_hosts(hosts)
+        self.randomize_hosts = randomize_hosts
+        self.hosts, chroot = collect_hosts(hosts, randomize_hosts)
         if chroot:
             self.chroot = normpath(chroot)
         else:
@@ -129,7 +184,6 @@ class KazooClient(object):
 
         # Curator like simplified state tracking, and listeners for
         # state transitions
-        self._state_lock = self.handler.rlock_object()
         self._state = KeeperState.CLOSED
         self.state = KazooState.LOST
         self.state_listeners = set()
@@ -141,8 +195,7 @@ class KazooClient(object):
             self._session_id = client_id[0]
             self._session_passwd = client_id[1]
         else:
-            self._session_id = None
-            self._session_passwd = str(bytearray([0] * 16))
+            self._reset_session()
 
         # ZK uses milliseconds
         self._session_timeout = int(timeout * 1000)
@@ -155,51 +208,86 @@ class KazooClient(object):
         self._stopped.set()
         self._writer_stopped.set()
 
-        self.retry = KazooRetry(
-            max_tries=max_retries,
-            delay=retry_delay,
-            backoff=retry_backoff,
-            max_jitter=retry_jitter,
-            sleep_func=self.handler.sleep_func
-        )
-        self.retry_sleeper = self.retry.retry_sleeper.copy()
+        self.retry = self._conn_retry = None
 
-        self._connection = ConnectionHandler(
-            self, self.retry.retry_sleeper.copy(), log_debug=self.log_debug)
+        if connection_retry is not None:
+            self._conn_retry = connection_retry
+            if self.handler.sleep_func != self._conn_retry.sleep_func:
+                raise ConfigurationError("Retry handler and event handler "
+                                         " must use the same sleep func")
 
-        # convenience API
-        from kazoo.recipe.barrier import Barrier
-        from kazoo.recipe.barrier import DoubleBarrier
-        from kazoo.recipe.election import Election
-        from kazoo.recipe.lock import Lock
-        from kazoo.recipe.partitioner import SetPartitioner
-        from kazoo.recipe.party import Party
-        from kazoo.recipe.party import ShallowParty
-        from kazoo.recipe.watchers import ChildrenWatch
-        from kazoo.recipe.watchers import DataWatch
+        if command_retry is not None:
+            self.retry = command_retry
+            if self.handler.sleep_func != self.comand_retry.sleep_func:
+                raise ConfigurationError("Command retry handler and event handler "
+                                         " must use the same sleep func")
+
+        if self.retry is None or self._conn_retry is None:
+            old_retry_keys = dict(_RETRY_COMPAT_DEFAULTS)
+            for key in old_retry_keys:
+                try:
+                    old_retry_keys[key] = kwargs.pop(key)
+                    warnings.warn('Passing retry configuration param %s to the'
+                            ' client directly is deprecated, please pass a'
+                            ' configured retry object (using param %s)' % (
+                                key, _RETRY_COMPAT_MAPPING[key]),
+                            DeprecationWarning, stacklevel=2)
+                except KeyError:
+                    pass
+
+            retry_keys = {}
+            for oldname, value in old_retry_keys.items():
+                retry_keys[_RETRY_COMPAT_MAPPING[oldname]] = value
+
+            if self._conn_retry is None:
+                self._conn_retry = KazooRetry(
+                    sleep_func=self.handler.sleep_func,
+                    **retry_keys)
+            if self.retry is None:
+                self.retry = KazooRetry(
+                    sleep_func=self.handler.sleep_func,
+                    **retry_keys)
+
+        self._conn_retry.interrupt = lambda: self._stopped.is_set()
+        self._connection = ConnectionHandler(self, self._conn_retry.copy(),
+            logger=self.logger)
 
         self.Barrier = partial(Barrier, self)
+        self.Counter = partial(Counter, self)
         self.DoubleBarrier = partial(DoubleBarrier, self)
         self.ChildrenWatch = partial(ChildrenWatch, self)
         self.DataWatch = partial(DataWatch, self)
         self.Election = partial(Election, self)
         self.Lock = partial(Lock, self)
         self.Party = partial(Party, self)
+        self.Queue = partial(Queue, self)
+        self.LockingQueue = partial(LockingQueue, self)
         self.SetPartitioner = partial(SetPartitioner, self)
+        self.Semaphore = partial(Semaphore, self)
         self.ShallowParty = partial(ShallowParty, self)
 
+         # If we got any unhandled keywords, complain like python would
+        if kwargs:
+            raise TypeError('__init__() got unexpected keyword arguments: %s'
+                            % (kwargs.keys(),))
+
     def _reset(self):
         """Resets a variety of client states for a new connection."""
-        with self._state_lock:
-            self._queue = self.handler.peekable_queue()
-            self._pending = self.handler.peekable_queue()
-            self._child_watchers = defaultdict(list)
-            self._data_watchers = defaultdict(list)
+        self._queue = deque()
+        self._pending = deque()
 
-        self._session_id = None
-        self._session_passwd = str(bytearray([0] * 16))
+        self._reset_watchers()
+        self._reset_session()
         self.last_zxid = 0
 
+    def _reset_watchers(self):
+        self._child_watchers = defaultdict(set)
+        self._data_watchers = defaultdict(set)
+
+    def _reset_session(self):
+        self._session_id = None
+        self._session_passwd = b'\x00' * 16
+
     @property
     def client_state(self):
         """Returns the last Zookeeper client state
@@ -267,14 +355,14 @@ class KazooClient(object):
                 if remove is True:
                     self.remove_listener(listener)
             except Exception:
-                log.exception("Error in connection state listener")
+                self.logger.exception("Error in connection state listener")
 
     def _session_callback(self, state):
         if state == self._state:
             return
 
-        # Note that we don't check self.state == LOST since thats also
-        # the clients initial state
+        # Note that we don't check self.state == LOST since that's also
+        # the client's initial state
         dead_state = self._state in LOST_STATES
         self._state = state
 
@@ -283,24 +371,26 @@ class KazooClient(object):
         # transitions since they only apply after
         # we've established a connection
         if dead_state and state == KeeperState.CONNECTING:
+            self.logger.debug("Skipping state change")
             return
 
-        if state == KeeperState.CONNECTED:
-            log.info("Zookeeper connection established")
+        if state in (KeeperState.CONNECTED, KeeperState.CONNECTED_RO):
+            self.logger.info("Zookeeper connection established, state: %s", state)
             self._live.set()
             self._make_state_change(KazooState.CONNECTED)
         elif state in LOST_STATES:
-            log.info("Zookeeper session lost, state: %s", state)
+            self.logger.info("Zookeeper session lost, state: %s", state)
             self._live.clear()
             self._make_state_change(KazooState.LOST)
             self._notify_pending(state)
             self._reset()
         else:
-            log.info("Zookeeper connection lost")
+            self.logger.info("Zookeeper connection lost")
             # Connection lost
             self._live.clear()
             self._notify_pending(state)
             self._make_state_change(KazooState.SUSPENDED)
+            self._reset_watchers()
 
     def _notify_pending(self, state):
         """Used to clear a pending response queue and request queue
@@ -311,33 +401,61 @@ class KazooClient(object):
             exc = SessionExpiredError()
         else:
             exc = ConnectionLoss()
-        while not self._pending.empty():
-            request, async_object, xid = self._pending.get()
-            if async_object:
-                async_object.set_exception(exc)
-        while not self._queue.empty():
-            request, async_object = self._queue.get()
-            if async_object:
-                async_object.set_exception(exc)
+
+        while True:
+            try:
+                request, async_object, xid = self._pending.popleft()
+                if async_object:
+                    async_object.set_exception(exc)
+            except IndexError:
+                break
+
+        while True:
+            try:
+                request, async_object = self._queue.popleft()
+                if async_object:
+                    async_object.set_exception(exc)
+            except IndexError:
+                break
 
     def _safe_close(self):
         self.handler.stop()
-        if not self._connection.stop(10):
-            raise Exception("Writer still open from prior connection"
-                            " and wouldn't close after 10 seconds")
+        timeout = self._session_timeout // 1000
+        if timeout < 10:
+            timeout = 10
+        if not self._connection.stop(timeout):
+            raise WriterNotClosedException(
+                "Writer still open from prior connection "
+                "and wouldn't close after %s seconds" % timeout)
 
     def _call(self, request, async_object):
         """Ensure there's an active connection and put the request in
         the queue if there is."""
-        with self._state_lock:
-            if self._state == KeeperState.AUTH_FAILED:
-                raise AuthFailedError()
-            elif self._state == KeeperState.CLOSED:
-                raise ConnectionClosedError("Connection has been closed")
-            elif self._state in (KeeperState.EXPIRED_SESSION,
-                                 KeeperState.CONNECTING):
-                raise SessionExpiredError()
-            self._queue.put((request, async_object))
+
+        if self._state == KeeperState.AUTH_FAILED:
+            async_object.set_exception(AuthFailedError())
+            return
+        elif self._state == KeeperState.CLOSED:
+            async_object.set_exception(ConnectionClosedError(
+                "Connection has been closed"))
+            return
+        elif self._state in (KeeperState.EXPIRED_SESSION,
+                             KeeperState.CONNECTING):
+            async_object.set_exception(SessionExpiredError())
+            return
+
+        self._queue.append((request, async_object))
+
+        # wake the connection, guarding against a race with close()
+        write_pipe = self._connection._write_pipe
+        if write_pipe is None:
+            async_object.set_exception(ConnectionClosedError(
+                "Connection has been closed"))
+        try:
+            os.write(write_pipe, b'\0')
+        except:
+            async_object.set_exception(ConnectionClosedError(
+                "Connection has been closed"))
 
     def start(self, timeout=15):
         """Initiate connection to ZK.
@@ -356,6 +474,10 @@ class KazooClient(object):
             self.stop()
             raise self.handler.timeout_exception("Connection time-out")
 
+        if self.chroot and not self.exists("/"):
+            warnings.warn("No chroot path exists, the chroot path "
+                          "should be created before normal use.")
+
     def start_async(self):
         """Asynchronously initiate connection to ZK.
 
@@ -399,7 +521,8 @@ class KazooClient(object):
             return
 
         self._stopped.set()
-        self._queue.put((Close(), None))
+        self._queue.append((CloseInstance, None))
+        os.write(self._connection._write_pipe, b'\0')
         self._safe_close()
 
     def restart(self):
@@ -407,13 +530,23 @@ class KazooClient(object):
         self.stop()
         self.start()
 
+    def close(self):
+        """Free any resources held by the client.
+
+        This method should be called on a stopped client before it is
+        discarded. Not doing so may result in filehandles being leaked.
+
+        .. versionadded:: 1.0
+        """
+        self._connection.close()
+
     def command(self, cmd=b'ruok'):
         """Sent a management command to the current ZK server.
 
         Examples are `ruok`, `envi` or `stat`.
 
         :returns: An unstructured textual response.
-        :rtype: unicode
+        :rtype: str
 
         :raises:
             :exc:`ConnectionLoss` if there is no connection open, or
@@ -507,7 +640,7 @@ class KazooClient(object):
         """
         return self.sync_async(path).get()
 
-    def create(self, path, value="", acl=None, ephemeral=False,
+    def create(self, path, value=b"", acl=None, ephemeral=False,
                sequence=False, makepath=False):
         """Create a node with the given value as its data. Optionally
         set an ACL on the node.
@@ -576,35 +709,19 @@ class KazooClient(object):
             returns a non-zero error code.
 
         """
-        try:
-            realpath = self.create_async(path, value, acl=acl,
-                ephemeral=ephemeral, sequence=sequence).get()
+        return self.create_async(path, value, acl=acl, ephemeral=ephemeral,
+            sequence=sequence, makepath=makepath).get()
 
-        except NoNodeError:
-            # some or all of the parent path doesn't exist. if makepath is set
-            # we will create it and retry. If it fails again, someone must be
-            # actively deleting ZNodes and we'd best bail out.
-            if not makepath:
-                raise
-
-            parent, _ = split(path)
-
-            # using the inner call directly because path is already namespaced
-            self._inner_ensure_path(parent, acl)
-
-            # now retry
-            realpath = self.create_async(path, value, acl=acl,
-                ephemeral=ephemeral, sequence=sequence).get()
-
-        return self.unchroot(realpath)
-
-    def create_async(self, path, value="", acl=None, ephemeral=False,
-                     sequence=False):
+    def create_async(self, path, value=b"", acl=None, ephemeral=False,
+                     sequence=False, makepath=False):
         """Asynchronously create a ZNode. Takes the same arguments as
-        :meth:`create`, with the exception of `makepath`.
+        :meth:`create`.
 
         :rtype: :class:`~kazoo.interfaces.IAsyncResult`
 
+        .. versionadded:: 1.1
+            The makepath option.
+
         """
         if acl is None and self.default_acl:
             acl = self.default_acl
@@ -614,12 +731,14 @@ class KazooClient(object):
         if acl and (isinstance(acl, ACL) or
                     not isinstance(acl, (tuple, list))):
             raise TypeError("acl must be a tuple/list of ACL's")
-        if not isinstance(value, str):
+        if not isinstance(value, bytes):
             raise TypeError("value must be a byte string")
         if not isinstance(ephemeral, bool):
             raise TypeError("ephemeral must be a bool")
         if not isinstance(sequence, bool):
             raise TypeError("sequence must be a bool")
+        if not isinstance(makepath, bool):
+            raise TypeError("makepath must be a bool")
 
         flags = 0
         if ephemeral:
@@ -630,7 +749,34 @@ class KazooClient(object):
             acl = OPEN_ACL_UNSAFE
 
         async_result = self.handler.async_result()
-        self._call(Create(_prefix_root(self.chroot, path), value, acl, flags),
+
+        def do_create():
+            self._create_async_inner(path, value, acl, flags, trailing=sequence).rawlink(create_completion)
+
+        @capture_exceptions(async_result)
+        def retry_completion(result):
+            result.get()
+            do_create()
+
+        @wrap(async_result)
+        def create_completion(result):
+            try:
+                return self.unchroot(result.get())
+            except NoNodeError:
+                if not makepath:
+                    raise
+                if sequence and path.endswith('/'):
+                    parent = path.rstrip('/')
+                else:
+                    parent, _ = split(path)
+                self.ensure_path_async(parent, acl).rawlink(retry_completion)
+
+        do_create()
+        return async_result
+
+    def _create_async_inner(self, path, value, acl, flags, trailing=False):
+        async_result = self.handler.async_result()
+        self._call(Create(_prefix_root(self.chroot, path, trailing=trailing), value, acl, flags),
                    async_result)
         return async_result
 
@@ -641,24 +787,46 @@ class KazooClient(object):
         :param acl: Permissions for node.
 
         """
-        self._inner_ensure_path(path, acl)
+        return self.ensure_path_async(path, acl).get()
 
-    def _inner_ensure_path(self, path, acl):
-        if self.exists(path):
-            return
+    def ensure_path_async(self, path, acl=None):
+        """Recursively create a path asynchronously if it doesn't
+        exist. Takes the same arguments as :meth:`ensure_path`.
 
-        if acl is None and self.default_acl:
-            acl = self.default_acl
+        :rtype: :class:`~kazoo.interfaces.IAsyncResult`
 
-        parent, node = split(path)
+        .. versionadded:: 1.1
 
-        if node:
-            self._inner_ensure_path(parent, acl)
-        try:
-            self.create_async(path, "", acl=acl).get()
-        except NodeExistsError:
-            # someone else created the node. how sweet!
-            pass
+        """
+        acl = acl or self.default_acl
+        async_result = self.handler.async_result()
+
+        @wrap(async_result)
+        def create_completion(result):
+            try:
+                return result.get()
+            except NodeExistsError:
+                return True
+
+        @capture_exceptions(async_result)
+        def prepare_completion(next_path, result):
+            result.get()
+            self.create_async(next_path, acl=acl).rawlink(create_completion)
+
+        @wrap(async_result)
+        def exists_completion(path, result):
+            if result.get():
+                return True
+            parent, node = split(path)
+            if node:
+                self.ensure_path_async(parent, acl=acl).rawlink(
+                    partial(prepare_completion, path))
+            else:
+                self.create_async(path, acl=acl).rawlink(create_completion)
+
+        self.exists_async(path).rawlink(partial(exists_completion, path))
+
+        return async_result
 
     def exists(self, path, watch=None):
         """Check if a node exists.
@@ -881,7 +1049,7 @@ class KazooClient(object):
                    async_result)
         return async_result
 
-    def set(self, path, data, version=-1):
+    def set(self, path, value, version=-1):
         """Set the value of a node.
 
         If the version of the node being updated is newer than the
@@ -895,7 +1063,7 @@ class KazooClient(object):
         than this will cause a ZookeeperError to be raised.
 
         :param path: Path of node.
-        :param data: New data value.
+        :param value: New data value.
         :param version: Version of node being updated, or -1.
         :returns: Updated :class:`~kazoo.protocol.states.ZnodeStat` of
                   the node.
@@ -914,9 +1082,9 @@ class KazooClient(object):
             returns a non-zero error code.
 
         """
-        return self.set_async(path, data, version).get()
+        return self.set_async(path, value, version).get()
 
-    def set_async(self, path, data, version=-1):
+    def set_async(self, path, value, version=-1):
         """Set the value of a node. Takes the same arguments as
         :meth:`set`.
 
@@ -925,16 +1093,33 @@ class KazooClient(object):
         """
         if not isinstance(path, basestring):
             raise TypeError("path must be a string")
-        if not isinstance(data, basestring):
-            raise TypeError("data must be a string")
+        if not isinstance(value, bytes):
+            raise TypeError("value must be a byte string")
         if not isinstance(version, int):
             raise TypeError("version must be an int")
 
         async_result = self.handler.async_result()
-        self._call(SetData(_prefix_root(self.chroot, path), data, version),
+        self._call(SetData(_prefix_root(self.chroot, path), value, version),
                    async_result)
         return async_result
 
+    def transaction(self):
+        """Create and return a :class:`TransactionRequest` object
+
+        Creates a :class:`TransactionRequest` object. A Transaction can
+        consist of multiple operations which can be committed as a
+        single atomic unit. Either all of the operations will succeed
+        or none of them.
+
+        :returns: A TransactionRequest.
+        :rtype: :class:`TransactionRequest`
+
+        .. versionadded:: 0.6
+            Requires Zookeeper 3.4+
+
+        """
+        return TransactionRequest(self)
+
     def delete(self, path, version=-1, recursive=False):
         """Delete a node.
 
@@ -1008,3 +1193,135 @@ class KazooClient(object):
             self.delete(path)
         except NoNodeError:  # pragma: nocover
             pass
+
+
+class TransactionRequest(object):
+    """A Zookeeper Transaction Request
+
+    A Transaction provides a builder object that can be used to
+    construct and commit an atomic set of operations. The transaction
+    must be committed before its sent.
+
+    Transactions are not thread-safe and should not be accessed from
+    multiple threads at once.
+
+    .. versionadded:: 0.6
+        Requires Zookeeper 3.4+
+
+    """
+    def __init__(self, client):
+        self.client = client
+        self.operations = []
+        self.committed = False
+
+    def create(self, path, value=b"", acl=None, ephemeral=False,
+               sequence=False):
+        """Add a create ZNode to the transaction. Takes the same
+        arguments as :meth:`KazooClient.create`, with the exception
+        of `makepath`.
+
+        :returns: None
+
+        """
+        if acl is None and self.client.default_acl:
+            acl = self.client.default_acl
+
+        if not isinstance(path, basestring):
+            raise TypeError("path must be a string")
+        if acl and not isinstance(acl, (tuple, list)):
+            raise TypeError("acl must be a tuple/list of ACL's")
+        if not isinstance(value, bytes):
+            raise TypeError("value must be a byte string")
+        if not isinstance(ephemeral, bool):
+            raise TypeError("ephemeral must be a bool")
+        if not isinstance(sequence, bool):
+            raise TypeError("sequence must be a bool")
+
+        flags = 0
+        if ephemeral:
+            flags |= 1
+        if sequence:
+            flags |= 2
+        if acl is None:
+            acl = OPEN_ACL_UNSAFE
+
+        self._add(Create(_prefix_root(self.client.chroot, path), value, acl,
+                         flags), None)
+
+    def delete(self, path, version=-1):
+        """Add a delete ZNode to the transaction. Takes the same
+        arguments as :meth:`KazooClient.delete`, with the exception of
+        `recursive`.
+
+        """
+        if not isinstance(path, basestring):
+            raise TypeError("path must be a string")
+        if not isinstance(version, int):
+            raise TypeError("version must be an int")
+        self._add(Delete(_prefix_root(self.client.chroot, path), version))
+
+    def set_data(self, path, value, version=-1):
+        """Add a set ZNode value to the transaction. Takes the same
+        arguments as :meth:`KazooClient.set`.
+
+        """
+        if not isinstance(path, basestring):
+            raise TypeError("path must be a string")
+        if not isinstance(value, bytes):
+            raise TypeError("value must be a byte string")
+        if not isinstance(version, int):
+            raise TypeError("version must be an int")
+        self._add(SetData(_prefix_root(self.client.chroot, path), value,
+                  version))
+
+    def check(self, path, version):
+        """Add a Check Version to the transaction.
+
+        This command will fail and abort a transaction if the path
+        does not match the specified version.
+
+        """
+        if not isinstance(path, basestring):
+            raise TypeError("path must be a string")
+        if not isinstance(version, int):
+            raise TypeError("version must be an int")
+        self._add(CheckVersion(_prefix_root(self.client.chroot, path),
+                  version))
+
+    def commit_async(self):
+        """Commit the transaction asynchronously.
+
+        :rtype: :class:`~kazoo.interfaces.IAsyncResult`
+
+        """
+        self._check_tx_state()
+        self.committed = True
+        async_object = self.client.handler.async_result()
+        self.client._call(Transaction(self.operations), async_object)
+        return async_object
+
+    def commit(self):
+        """Commit the transaction.
+
+        :returns: A list of the results for each operation in the
+                  transaction.
+
+        """
+        return self.commit_async().get()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, exc_tb):
+        """Commit and cleanup accumulated transaction data."""
+        if not exc_type:
+            self.commit()
+
+    def _check_tx_state(self):
+        if self.committed:
+            raise ValueError('Transaction already committed')
+
+    def _add(self, request, post_processor=None):
+        self._check_tx_state()
+        self.client.logger.debug('Added %r to %r', request, self)
+        self.operations.append(request)
