@@ -26,7 +26,21 @@ class ZookeeperStoppedError(KazooException):
 
 
 class ConnectionDropped(KazooException):
-    """ Internal error for jumping out of loops """
+    """Internal error for jumping out of loops"""
+
+
+class LockTimeout(KazooException):
+    """Raised if failed to acquire a lock.
+
+    .. versionadded:: 1.1
+    """
+
+
+class WriterNotClosedException(KazooException):
+    """Raised if the writer is unable to stop closing when requested.
+
+    .. versionadded:: 1.2
+    """
 
 
 def _invalid_error_code():
@@ -148,6 +162,17 @@ class AuthFailedError(ZookeeperError):
     pass
 
 
+@_zookeeper_exception(-118)
+class SessionMovedError(ZookeeperError):
+    pass
+
+
+@_zookeeper_exception(-119)
+class NotReadOnlyCallError(ZookeeperError):
+    """An API call that is not read-only was used while connected to
+    a read-only server"""
+
+
 class ConnectionClosedError(SessionExpiredError):
     """Connection is closed"""
 
