@@ -1,10 +1,10 @@
 #
 #  Copyright (C) 2010-2011, 2011 Canonical Ltd. All Rights Reserved
 #
-#  This file is part of txzookeeper.
+#  This file was originally taken from txzookeeper and modified later.
 #
 #  Authors:
-#   Kapil Thangavelu
+#   Kapil Thangavelu and the Kazoo team
 #
 #  txzookeeper is free software: you can redistribute it and/or modify
 #  it under the terms of the GNU Lesser General Public License as published by
@@ -18,20 +18,46 @@
 #
 #  You should have received a copy of the GNU Lesser General Public License
 #  along with txzookeeper.  If not, see <http://www.gnu.org/licenses/>.
-#
 
 
+import code
 import os
 import os.path
 import shutil
+import signal
 import subprocess
 import tempfile
+import traceback
 
 from itertools import chain
 from collections import namedtuple
 from glob import glob
 
 
+def debug(sig, frame):
+    """Interrupt running process, and provide a python prompt for
+    interactive debugging."""
+    d = {'_frame': frame}         # Allow access to frame object.
+    d.update(frame.f_globals)  # Unless shadowed by global
+    d.update(frame.f_locals)
+
+    i = code.InteractiveConsole(d)
+    message = "Signal recieved : entering python shell.\nTraceback:\n"
+    message += ''.join(traceback.format_stack(frame))
+    i.interact(message)
+
+
+def listen():
+    if os.name != 'nt':  # SIGUSR1 is not supported on Windows
+        signal.signal(signal.SIGUSR1, debug)  # Register handler
+listen()
+
+
+def to_java_compatible_path(path):
+    if os.name == 'nt':
+        path = path.replace('\\', '/')
+    return path
+
 ServerInfo = namedtuple(
     "ServerInfo", "server_id client_port election_port leader_port")
 
@@ -44,13 +70,14 @@ class ManagedZooKeeper(object):
     future, we may want to do that, especially when run in a
     Hudson/Buildbot context, to ensure more test robustness."""
 
-    def __init__(self, software_path, server_info, peers=()):
+    def __init__(self, software_path, server_info, peers=(), classpath=None):
         """Define the ZooKeeper test instance.
 
         @param install_path: The path to the install for ZK
         @param port: The port to run the managed ZK instance
         """
         self.install_path = software_path
+        self._classpath = classpath
         self.server_info = server_info
         self.host = "127.0.0.1"
         self.peers = peers
@@ -83,7 +110,7 @@ tickTime=2000
 dataDir=%s
 clientPort=%s
 maxClientCnxns=0
-""" % (data_path, self.server_info.client_port))
+""" % (to_java_compatible_path(data_path), self.server_info.client_port))
 
         # setup a replicated setup if peers are specified
         if self.peers:
@@ -111,7 +138,7 @@ log4j.appender.ROLLINGFILE.layout=org.apache.log4j.PatternLayout
 log4j.appender.ROLLINGFILE.layout.ConversionPattern=%d{ISO8601} [myid:%X{myid}] - %-5p [%t:%C{1}@%L] - %m%n
 log4j.appender.ROLLINGFILE=org.apache.log4j.RollingFileAppender
 log4j.appender.ROLLINGFILE.Threshold=DEBUG
-log4j.appender.ROLLINGFILE.File=""" + (
+log4j.appender.ROLLINGFILE.File=""" + to_java_compatible_path(
                 self.working_path + os.sep + "zookeeper.log\n"))
 
         self.process = subprocess.Popen(
@@ -123,14 +150,16 @@ log4j.appender.ROLLINGFILE.File=""" + (
                   "-Dlog4j.configuration=file:%s" % log4j_path,
                   # "-Dlog4j.debug",
                   "org.apache.zookeeper.server.quorum.QuorumPeerMain",
-                  config_path],
-            )
+                  config_path])
         self._running = True
 
     @property
     def classpath(self):
         """Get the classpath necessary to run ZooKeeper."""
 
+        if self._classpath:
+            return self._classpath
+
         # Two possibilities, as seen in zkEnv.sh:
         # Check for a release - top-level zookeeper-*.jar?
         jars = glob((os.path.join(
@@ -140,6 +169,13 @@ log4j.appender.ROLLINGFILE.File=""" + (
             jars.extend(glob(os.path.join(
                 self.install_path,
                 "lib/*.jar")))
+            # support for different file locations on Debian/Ubuntu
+            jars.extend(glob(os.path.join(
+                self.install_path,
+                "log4j-*.jar")))
+            jars.extend(glob(os.path.join(
+                self.install_path,
+                "slf4j-*.jar")))
         else:
             # Development build (plain `ant`)
             jars = glob((os.path.join(
@@ -147,7 +183,8 @@ log4j.appender.ROLLINGFILE.File=""" + (
             jars.extend(glob(os.path.join(
                 self.install_path,
                 "build/lib/*.jar")))
-        return ":".join(jars)
+
+        return os.pathsep.join(jars)
 
     @property
     def address(self):
@@ -189,8 +226,9 @@ log4j.appender.ROLLINGFILE.File=""" + (
 
 class ZookeeperCluster(object):
 
-    def __init__(self, install_path, size=3, port_offset=20000):
+    def __init__(self, install_path=None, classpath=None, size=3, port_offset=20000):
         self._install_path = install_path
+        self._classpath = classpath
         self._servers = []
 
         # Calculate ports and peer group
@@ -198,9 +236,9 @@ class ZookeeperCluster(object):
         peers = []
 
         for i in range(size):
-            port += i * 10
             info = ServerInfo(i + 1, port, port + 1, port + 2)
             peers.append(info)
+            port += 10
 
         # Instantiate Managed ZK Servers
         for i in range(size):
@@ -208,7 +246,7 @@ class ZookeeperCluster(object):
             server_info = server_peers.pop(i)
             self._servers.append(
                 ManagedZooKeeper(
-                    self._install_path, server_info, server_peers))
+                    self._install_path, server_info, server_peers, classpath=self._classpath))
 
     def __getitem__(self, k):
         return self._servers[k]
