@@ -1,22 +1,26 @@
-import atexit
-import os
 import sys
 import threading
 import time
 import uuid
 import unittest
 
+from mock import patch
 from nose import SkipTest
 from nose.tools import eq_
 from nose.tools import raises
 
 from kazoo.testing import KazooTestCase
-from kazoo.exceptions import BadArgumentsError
-from kazoo.exceptions import ConfigurationError
-from kazoo.exceptions import InvalidACLError
-from kazoo.exceptions import NoNodeError
-from kazoo.exceptions import NoAuthError
-from kazoo.exceptions import ConnectionLoss
+from kazoo.exceptions import (
+    BadArgumentsError,
+    ConfigurationError,
+    ConnectionClosedError,
+    ConnectionLoss,
+    InvalidACLError,
+    NoAuthError,
+    NoNodeError,
+    NodeExistsError,
+)
+
 
 if sys.version_info > (3, ):  # pragma: nocover
     def u(s):
@@ -26,48 +30,6 @@ else:  # pragma: nocover
         return unicode(s, "unicode_escape")
 
 
-class TestSessions(unittest.TestCase):
-
-    def setUp(self):
-        from kazoo.client import KazooClient
-        from kazoo.protocol.states import KazooState
-        from kazoo.testing.common import ZookeeperCluster
-        ZK_HOME = os.environ.get("ZOOKEEPER_PATH")
-        self.cluster = ZookeeperCluster(ZK_HOME, size=1, port_offset=21000)
-        self.cluster.start()
-        atexit.register(lambda cluster: self.cluster.terminate(), self.cluster)
-        self.client = KazooClient(self.cluster[0].address, max_retries=3)
-        self.ev = threading.Event()
-
-        def back(state):
-            if state == KazooState.CONNECTED:
-                self.ev.set()
-        self.client.start()
-        self.path = self.client.create(uuid.uuid4().hex)
-        self.client.add_listener(back)
-
-    def test_restarted_server(self):
-        self.cluster.stop()
-        self.cluster.start()
-        self.ev.wait(5)
-        raise SkipTest('Patch missing')
-        eq_(self.ev.is_set(), True)
-        self.assertTrue(self.client.retry(self.client.exists, self.path))
-
-    def test_terminated_server(self):
-        self.cluster.reset()
-        self.cluster.start()
-        self.ev.wait(5)
-        raise SkipTest('Patch missing')
-        eq_(self.ev.is_set(), True)
-        self.assertFalse(self.client.retry(self.client.exists, self.path))
-
-    def tearDown(self):
-        self.ev.clear()
-        self.client.stop()
-        self.cluster.stop()
-
-
 class TestClientTransitions(KazooTestCase):
     def test_connection_and_disconnection(self):
         from kazoo.client import KazooState
@@ -91,8 +53,9 @@ class TestClientTransitions(KazooTestCase):
         states.pop()
         self.expire_session()
         rc.wait(2)
-        eq_(states, [KazooState.SUSPENDED, KazooState.LOST,
-                     KazooState.CONNECTED])
+
+        req_states = [KazooState.LOST, KazooState.CONNECTED]
+        eq_(states, req_states)
 
 
 class TestClientConstructor(unittest.TestCase):
@@ -137,6 +100,16 @@ class TestConnection(KazooTestCase):
         from kazoo.security import make_digest_acl
         return make_digest_acl(*args, **kwargs)
 
+    def test_chroot_warning(self):
+        k = self._get_nonchroot_client()
+        k.chroot = 'abba'
+        try:
+            with patch('warnings.warn') as mock_func:
+                k.start()
+                assert mock_func.called
+        finally:
+            k.stop()
+
     def test_auth(self):
         username = uuid.uuid4().hex
         password = uuid.uuid4().hex
@@ -164,6 +137,8 @@ class TestConnection(KazooTestCase):
         finally:
             # Ensure we remove the ACL protected nodes
             self.client.delete("/1", recursive=True)
+            eve.stop()
+            eve.close()
 
     def test_connect_auth(self):
         username = uuid.uuid4().hex
@@ -182,6 +157,7 @@ class TestConnection(KazooTestCase):
         finally:
             client.delete('/1')
             client.stop()
+            client.close()
 
     def test_unicode_auth(self):
         username = u("xe4/\hm")
@@ -208,6 +184,8 @@ class TestConnection(KazooTestCase):
         finally:
             # Ensure we remove the ACL protected nodes
             self.client.delete("/1", recursive=True)
+            eve.stop()
+            eve.close()
 
     def test_invalid_auth(self):
         self.assertRaises(TypeError, self.client.add_auth,
@@ -312,6 +290,36 @@ class TestConnection(KazooTestCase):
         self.client.restart()
         self.assertTrue(self.client.connected)
 
+    def test_closed(self):
+        client = self.client
+        client.stop()
+
+        write_pipe = client._connection._write_pipe
+
+        # close the connection to free the pipe
+        client.close()
+        eq_(client._connection._write_pipe, None)
+
+        from kazoo.protocol.states import KeeperState
+
+        # sneak in and patch client to simulate race between a thread
+        # calling stop(); close() and one running a command
+        oldstate = client._state
+        client._state = KeeperState.CONNECTED
+        client._connection._write_pipe = write_pipe
+        try:
+            # simulate call made after write pipe is closed
+            self.assertRaises(ConnectionClosedError, client.exists, '/')
+
+            # simualte call made after write pipe is set to None
+            client._connection._write_pipe = None
+            self.assertRaises(ConnectionClosedError, client.exists, '/')
+
+        finally:
+            # reset for teardown
+            client._state = oldstate
+            client._connection._write_pipe = None
+
 
 class TestClient(KazooTestCase):
     def _getKazooState(self):
@@ -341,6 +349,11 @@ class TestClient(KazooTestCase):
         path = client.create(u("/\xe4hm"))
         eq_(path, u("/\xe4hm"))
 
+    def test_create_async_returns_unchrooted_path(self):
+        client = self.client
+        path = client.create_async('/1').get()
+        eq_(path, "/1")
+
     def test_create_invalid_path(self):
         client = self.client
         self.assertRaises(TypeError, client.create, ('a', ))
@@ -358,6 +371,7 @@ class TestClient(KazooTestCase):
         self.assertRaises(TypeError, client.create, 'a', value=['a'])
         self.assertRaises(TypeError, client.create, 'a', ephemeral='yes')
         self.assertRaises(TypeError, client.create, 'a', sequence='yes')
+        self.assertRaises(TypeError, client.create, 'a', makepath='yes')
 
     def test_create_value(self):
         client = self.client
@@ -423,6 +437,8 @@ class TestClient(KazooTestCase):
         eq_(path, "/folder/a0000000000")
         path2 = client.create("/folder/a", b"sequence", sequence=True)
         eq_(path2, "/folder/a0000000001")
+        path3 = client.create("/folder/", b"sequence", sequence=True)
+        eq_(path3, "/folder/0000000002")
 
     def test_create_ephemeral_sequence(self):
         basepath = "/" + uuid.uuid4().hex
@@ -441,11 +457,34 @@ class TestClient(KazooTestCase):
         data, stat = self.client.get("/1/2/3/4/5")
         eq_(data, b"val2")
 
+        self.assertRaises(NodeExistsError, self.client.create, "/1/2/3/4/5",
+            b"val2", makepath=True)
+
+    def test_create_makepath_incompatible_acls(self):
+        from kazoo.client import KazooClient
+        from kazoo.security import make_digest_acl_credential, CREATOR_ALL_ACL
+        credential = make_digest_acl_credential("username", "password")
+        alt_client = KazooClient(self.cluster[0].address + self.client.chroot,
+            max_retries=5, auth_data=[("digest", credential)])
+        alt_client.start()
+        alt_client.create("/1/2", b"val2", makepath=True, acl=CREATOR_ALL_ACL)
+
+        try:
+            self.assertRaises(NoAuthError, self.client.create, "/1/2/3/4/5",
+                b"val2", makepath=True)
+        finally:
+            alt_client.delete('/', recursive=True)
+            alt_client.stop()
+
     def test_create_no_makepath(self):
         self.assertRaises(NoNodeError, self.client.create, "/1/2", b"val1")
         self.assertRaises(NoNodeError, self.client.create, "/1/2", b"val1",
             makepath=False)
 
+        self.client.create("/1/2", b"val1", makepath=True)
+        self.assertRaises(NoNodeError, self.client.create, "/1/2/3/4", b"val1",
+            makepath=False)
+
     def test_create_exists(self):
         from kazoo.exceptions import NodeExistsError
         client = self.client
@@ -699,14 +738,6 @@ class TestClient(KazooTestCase):
             client.get('/')
         testit()
 
-    def test_bad_safeclose(self):
-        client = self.client
-
-        @raises(Exception)
-        def testit():
-            client._safe_close()
-        testit()
-
     def test_client_state(self):
         from kazoo.protocol.states import KeeperState
         eq_(self.client.client_state, KeeperState.CONNECTED)
