@@ -1,10 +1,10 @@
 import uuid
 import time
 
-
 from nose.tools import eq_
 
 from kazoo.testing import KazooTestCase
+from kazoo.recipe.partitioner import PartitionState
 
 
 class KazooPartitionerTests(KazooTestCase):
@@ -15,9 +15,8 @@ class KazooPartitionerTests(KazooTestCase):
     def test_party_of_one(self):
         partitioner = self.client.SetPartitioner(
             self.path, set=(1, 2, 3), time_boundary=0.2)
-        eq_(partitioner.acquired, False)
-        partitioner.wait_for_acquire(1)
-        eq_(partitioner.acquired, True)
+        partitioner.wait_for_acquire(14)
+        eq_(partitioner.state, PartitionState.ACQUIRED)
         eq_(list(partitioner), [1, 2, 3])
         partitioner.finish()
 
@@ -26,9 +25,8 @@ class KazooPartitionerTests(KazooTestCase):
                         identifier="p%s" % i, time_boundary=0.2)
                         for i in range(2)]
 
-        eq_(partitioners[0].acquired, False)
-        partitioners[0].wait_for_acquire(1)
-        partitioners[1].wait_for_acquire(1)
+        partitioners[0].wait_for_acquire(14)
+        partitioners[1].wait_for_acquire(14)
         eq_(list(partitioners[0]), [1])
         eq_(list(partitioners[1]), [2])
         partitioners[0].finish()
@@ -41,9 +39,10 @@ class KazooPartitionerTests(KazooTestCase):
                         identifier="p%s" % i, time_boundary=0.2)
                         for i in range(2)]
 
-        eq_(partitioners[0].acquired, False)
-        partitioners[0].wait_for_acquire(1)
-        partitioners[1].wait_for_acquire(1)
+        partitioners[0].wait_for_acquire(14)
+        partitioners[1].wait_for_acquire(14)
+        eq_(partitioners[0].state, PartitionState.ACQUIRED)
+        eq_(partitioners[1].state, PartitionState.ACQUIRED)
 
         eq_(list(partitioners[0]), [1, 3])
         eq_(list(partitioners[1]), [2])
@@ -57,7 +56,7 @@ class KazooPartitionerTests(KazooTestCase):
             p.release_set()
 
         for p in partitioners:
-            p.wait_for_acquire(1)
+            p.wait_for_acquire(14)
 
         eq_(list(partitioners[0]), [1])
         eq_(list(partitioners[1]), [2])
@@ -71,12 +70,23 @@ class KazooPartitionerTests(KazooTestCase):
                         identifier="p%s" % i, time_boundary=0.2)
                         for i in range(2)]
 
-        eq_(partitioners[0].acquired, False)
-        partitioners[0].wait_for_acquire(1)
-        partitioners[1].wait_for_acquire(1)
+        partitioners[0].wait_for_acquire(14)
+        partitioners[1].wait_for_acquire(14)
+        eq_(partitioners[0].state, PartitionState.ACQUIRED)
+        eq_(partitioners[1].state, PartitionState.ACQUIRED)
 
         eq_(list(partitioners[0]), [1])
         eq_(list(partitioners[1]), [])
 
         for p in partitioners:
             p.finish()
+
+    def test_party_session_failure(self):
+        partitioner = self.client.SetPartitioner(
+            self.path, set=(1, 2, 3), time_boundary=0.2)
+        partitioner.wait_for_acquire(14)
+        eq_(partitioner.state, PartitionState.ACQUIRED)
+        # simulate session failure
+        partitioner._fail_out()
+        partitioner.release_set()
+        self.assertTrue(partitioner.failed)
